'''
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
'''

import logging
from typing import Dict

from ..exceptions import CertRecordParseException, RemoteCertificateExpired, TofuCertificateRejection
from .cert_record import CertRecord
from .cert_wrapper import CertWrapper

logger = logging.getLogger(__name__)


class CertStore:
  '''
  Data structure to store the certificates across visited hosts
  '''

  __hosts_file: str
  __cert_store_data: Dict[str, CertRecord]

  def __init__(self, hosts_file):
    '''
    Initializes a new cert store with a specified file to store the certificate fingerprint & expiration dates
    '''
    self.__cert_store_data = {}
    self.__hosts_file = hosts_file

  def set_hosts_file(self, hosts_file):
    '''
    Updates the specified file for certificate fingerprint storage
    '''
    self.__hosts_file = hosts_file

  def get_hosts_file(self):
    '''
    Returns the currently set hosts file location
    '''
    return self.__hosts_file

  def validate_tofu_or_add(self, hostname: str, cert: CertWrapper) -> bool:
    '''
    Given the hostname & correspoding certificate, this function:
    1. Checks to see if the certificate is expired (if so, it throws a RemoteCertificateExpired exception)
    2. Fetches a corresponding stored certificate record from the client to implement a TOFU check:
      a. If there is a local cert record, and it's not expired, and it matches the passed certificate, return success
      b. If there is not a local cert record, save the certificate record locally, and return success
      c. If there is a local cert record, but it's expired, save the certificate record locally, and return success
      d. If there is a local cert record, and it's not expired, but it does not match the passed certificate, throw TofuCertificateRejection
    '''
    remote_cert_record = CertRecord(hostname, cert.fingerprint(), cert.expiration())

    if remote_cert_record.is_expired():
      raise RemoteCertificateExpired

    local_cert_record = self.__get_cert_record(hostname)

    if local_cert_record and not local_cert_record.is_expired() and local_cert_record.fingerprint != remote_cert_record.fingerprint:
      raise TofuCertificateRejection

    self.__add_cert_record(remote_cert_record)
    return True

  def __get_cert_record(self, hostname: str) -> CertRecord:
    '''
    Fetch the corresponding CertRecord for passed hostname from the local storage (file)
    TODO: smarter loading logic
    '''
    self.__load()
    return self.__cert_store_data.get(hostname, None)

  def __add_cert_record(self, cert_record: CertRecord):
    '''
    Add a CertRecord for the corresponding hostname to local storage (file) and save to file
    TODO: smarter saving logic
    '''
    self.__cert_store_data[cert_record.hostname] = cert_record
    self.__save()
    return self

  def __load(self):
    '''
    Reloads the hosts file from storage and copies that into memory
    '''
    file_lines = []
    try:
      with open(self.__hosts_file, 'r') as f:
        file_lines = f.readlines()
    except FileNotFoundError:
      file_lines = []

    for file_line in file_lines:
      cert_record = self.__load_record(file_line)
      if cert_record is not None:
        self.__cert_store_data[cert_record.hostname] = cert_record

    return self

  def __load_record(self, file_line):
    try:
      return CertRecord.from_string(file_line)
    except CertRecordParseException:
      logger.warning(f"Invalid TOFU record encountered: '{file_line.strip()}'. This record has been skipped.")
      return None

  def __save(self):
    '''
    Saves the full set of host records back to file
    '''
    with open(self.__hosts_file, 'w') as f:
      for c in self.__cert_store_data.values():
        f.write(c.to_string())

    return self



'''
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
'''

import cryptography
from cryptography import x509
from cryptography.hazmat.backends import default_backend


class CertWrapper:
  '''
  Certificate as defined by x509
  '''
  certificate: cryptography.x509.Certificate
  '''
  Certificate fingerprint, to be used in TOFU handling and response
  '''
  public_key_fingerprint: str

  def __init__(self, certificate: cryptography.x509.Certificate):
    '''
    Constructor
    '''
    self.certificate = certificate

  def expiration(self) -> str:
    '''
    Access function for certificate expiration date
    '''
    return self.certificate.not_valid_after

  def fingerprint(self) -> str:
    '''
    Extracts the public key & expiration date from the cert,
    and returns the public key openssh fingerprint
    '''
    return self.certificate.public_key().public_bytes(
      cryptography.hazmat.primitives.serialization.Encoding.OpenSSH,
      cryptography.hazmat.primitives.serialization.PublicFormat.OpenSSH
    ).decode('utf-8')

  @classmethod
  def parse(cls, raw_certificate: bytes):
    '''
    Takes as input the raw certificate (originally from the TCP socket)
    Returns a certificate wrapper
    '''
    x509_certificate = x509.load_der_x509_certificate(raw_certificate, default_backend())
    return CertWrapper(x509_certificate)



'''
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
'''

import logging

logger = logging.getLogger(__name__)


def normalize_path(path: str) -> str:
  '''
  Implements a normalized path for a string
  See RFC-3986 https://tools.ietf.org/html/rfc3986#section-5.2.4
  5.2.4.  Remove Dot Segments

  Example:
    STEP   OUTPUT BUFFER         INPUT BUFFER

      1 :                         /a/b/c/./../../g
      2E:   /a                    /b/c/./../../g
      2E:   /a/b                  /c/./../../g
      2E:   /a/b/c                /./../../g
      2B:   /a/b/c                /../../g
      2C:   /a/b                  /../g
      2C:   /a                    /g
      2E:   /a/g
  '''
  result_stack = []
  for component in path.split('/'):
    if component in ('.', ''):
      continue  # Do nothing
    if component == '..':
      if len(result_stack) > 0:
        result_stack.pop()
    else:
      result_stack.append(component)

  unescaped_path = ''.join([
    ('/' if len(path) > 0 and path[0] == '/' else ''),
    ('/'.join(result_stack)),
    ('/' if len(path) > 0 and path[len(path) - 1] == '/' else ''),
  ])

  return unescaped_path.replace("//", "/")


class TimeoutManager:
  '''
  Timeout Manager for global timeout management at the top-level
  '''
  def __init__(self, default_timeout):
    '''
    Sets a default timeout on initialization
    '''
    self.set_default_timeout(default_timeout)

  def set_default_timeout(self, default_timeout):
    '''
    Allow the default timeout to be overwritten
    '''
    self.default_timeout = default_timeout

  def get_timeout(self, timeout):
    '''
    Takes in a timeout and returns that, or the default timeout
    '''
    if timeout is not None:
      return timeout
    return self.default_timeout



'''
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
'''

import cgi
import logging

from cryptography.x509 import Certificate

from .globals import *

logger = logging.getLogger(__name__)


class ResponseFactory:
  '''
  Wrapper class for factory:
  Determines the approriate response type based on response status
  and generates the appropriate response type
  '''
  @classmethod
  def create(cls, url: str, status: str, meta=None, raw_body=None, certificate=None):
    '''
    Given a url, status, and response data, generates the appropriate response type
    '''
    basic_status_code = status[0]
    factories = {
      "0": ErrorResponse,
      "1": InputResponse,
      "2": SuccessResponse,
      "3": RedirectResponse,
      "4": TempFailureResponse,
      "5": PermFailureResponse,
      "6": ClientCertRequiredResponse,
    }

    factory_class = factories.get(basic_status_code, None)

    if factory_class is None:
      return ErrorResponse(
        url,
        RESPONSE_STATUSDETAIL_ERROR_PROTOCOL,
        f"Invalid response received from the server, status code: {status}",
        None,
        None,
      )

    return factory_class(url, status, meta, raw_body, certificate)


class BaseResponse:
  '''
  Abstract Base response type that all response types inherit from.
  Included public members:
  * url
  * basic_status
  * status
  * meta
  * raw_body
  * certificate
  '''

  url: str
  basic_status: str
  status: str
  meta: str
  raw_body: bytes
  certificate: Certificate

  def __init__(self, url: str, status: str, meta: str, raw_body: bytes, certificate: Certificate):
    '''
    Initializes a BaseResponse with the request url, status code, metadata, raw body string, and remote certificate
    '''
    self.url = str(url)
    self.basic_status = status[0]
    self.status = status
    self.meta = meta
    self.raw_body = raw_body
    self.certificate = certificate

  def is_a(self, response_class_type):
    '''
    Returns true if the response class type matches the current class
    '''
    return isinstance(self, response_class_type)

  def success(self):
    '''
    Returns true if the response is of the type success
    '''
    return self.is_a(SuccessResponse)

  def data(self):
    '''
    Fetches processed data from the response.  This method should be overloaded in each specific response type.
    '''
    return self.raw_body

  def __str__(self):
    '''
    Returns a literal string representation of the response, including messageheader
    '''
    return f"{self.status} {self.meta}"

  def __repr__(self):
    '''
    A representation of the string for developers
    '''
    return f'<ignition.response.{self.__class__.__name__}: {self.status} {self.meta} [from {self.url}]>'


class ErrorResponse(BaseResponse):
  '''
  ErrorResponse
  This is a custom response type for ignition, to handle any responses representing request errors
  that are outside of the scope of the Gemini protocol.  Included options are:

  00: RESPONSE_STATUSDETAIL_ERROR_NETWORK
  Any errors that occur at the network level, and prevented the client from making any connection
  with external services.

  01: RESPONSE_STATUSDETAIL_ERROR_DNS = "01"
  Any errors at the DNS level.

  02: RESPONSE_STATUSDETAIL_ERROR_HOST
  Any errors connecting to the host (timeout, refused, etc.).

  03: RESPONSE_STATUSDETAIL_ERROR_TLS
  Any errors associated with TLS/SSL, including handshake errors, certificate expired errors,
  and security errors like certificate rejection errors.

  04: RESPONSE_STATUSDETAIL_ERROR_PROTOCOL
  Any errors where a secure message is received from the server, but it does not conform to the
  Gemini protocol requirements and cannot be processed.
  '''
  def data(self):
    '''
    Fetch data relevant to the ErrorResponse; in this case the metadata message from the response
    '''
    return self.meta


class InputResponse(BaseResponse):
  '''
  InputRequest
  Meets Gemini specification: 3.2.1 1x (INPUT)

  Status codes beginning with 1 are INPUT status codes, meaning that
  the requested resource accepts a line of textual user input.

  The user should reissue a request to the url with parameters in the form:
  gemini://hostname/path?query
  '''
  def data(self):
    '''
    Returns the related instructions for the InputResponse.
    The <META> line is a prompt which should be displayed to the user.
    '''
    return self.meta


class SuccessResponse(BaseResponse):
  '''
  SuccessResponse
  Meets Gemini specification: 3.2.2 2x (SUCCESS)

  Status codes beginning with 2 are SUCCESS status codes.
  '''
  def data(self):
    '''
    Decode the success message body using metadata in the appropriate encoding type
    '''

    meta = self.meta or GEMINI_DEFAULT_MIME_TYPE
    _, options = cgi.parse_header(meta)
    encoding = options['charset'] if 'charset' in options else GEMINI_DEFAULT_ENCODING
    try:
      return self.raw_body.decode(encoding)
    except LookupError:
      logger.warning(f"Could not decode response body using invalid encoding {encoding}")
      return self.raw_body

  def __str__(self):
    '''
    The string representation of the success message should be header + body
    '''
    return f"{self.status} {self.meta}{CRLF}{self.data()}"


class RedirectResponse(BaseResponse):
  '''
  RedirectResponse
  Meets Gemini specification: 3.2.3 3x (REDIRECT)

  Status codes beginning with 3 are REDIRECT status codes.

  The server is redirecting the client to a new location for the requested resource
  '''
  def data(self):
    '''
    Returns the new destination for redirection from the server
    '''
    return self.meta


class TempFailureResponse(BaseResponse):
  '''
  TempFailureResponse
  Meets Gemini specification: 3.2.4 4x (TEMPORARY FAILURE)

  Status codes beginning with 4 are TEMPORARY FAILURE status codes.

  The request has failed, but an identical request may success in the future.
  '''
  def data(self):
    '''
    Returns the data from the server in the META field, which may provide additional information to the user.
    '''
    return f"{self.status} {self.meta}"


class PermFailureResponse(BaseResponse):
  '''
  PermFailureResponse
  Meets Gemini specification: 3.2.5 5x (PERMANENT FAILURE)

  Status codes beginning with 5 are PERMANENT FAILURE status codes.

  The request has failed, identical requests will likely fail in the future.
  '''
  def data(self):
    '''
    Returns the data from the server in the META field, which may provide additional information to the user.
    '''
    return f"{self.status} {self.meta}"


class ClientCertRequiredResponse(BaseResponse):
  '''
  ClientCertRequiredResponse
  Meets Gemini specification: 3.2.6 6x (CLIENT CERTIFICATE REQUIRED)

  Status codes beginning with 6 are CLIENT CERTIFICATE REQUIRED status codes

  The request should be retried with a client certificate.
  '''
  def data(self):
    '''
    Return additional information from the server on certificate requirements
    or the reason a certificate was rejected
    '''
    return self.meta



'''
This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL
was not distributed with this file, You can obtain one
at http://mozilla.org/MPL/2.0/.
'''

from .globals import *
from .request import Request
from .response import (
  ClientCertRequiredResponse,
  ErrorResponse,
  InputResponse,
  PermFailureResponse,
  RedirectResponse,
  SuccessResponse,
  TempFailureResponse,
)
from .ssl.cert_store import CertStore
from .util import TimeoutManager

__version__ = '0.1.7'

__timeout = TimeoutManager(DEFAULT_REQUEST_TIMEOUT)
__cert_store = CertStore(DEFAULT_HOSTS_FILE)


def set_default_hosts_file(hosts_file):
  '''
  Set the default host file location where all of the certificate fingerprints
  are stored in order to support Trust-On-First-Use (TOFU) validation.
  By default, this file is stored in the root directory as your project in a
  file named `.known_hosts`. This can be updated to any readable location but
  should be stored somewhere persistent for security purposes.

  The format of this file is very similar to (but not identical to)
  the SSH `known_hosts` file.

  Parameters:
  * hosts_file: `string`

  '''
  __cert_store.set_hosts_file(hosts_file)


def set_default_timeout(timeout):
  '''
  Set the default timeout (in seconds) for all requests made via ignition.
  The default timeout is 30 seconds.

  Parameters:
  * timeout: `float`
  '''
  __timeout.set_default_timeout(timeout)


def url(request_url, referer=None):
  '''
  Given a *url* to a Gemini capsule, this returns a standardized,
  fully-qualified url to the Gemini capsule.  If a *referer* is
  provided, a dynamic URL is constructed by ignition to send a request
  to.  This logic follows URL definition behavior outlined in
  [RFC-3986](https://tools.ietf.org/html/rfc3986).

  This allows for the bulk of URL generation logic to be handled
  without ignition as opposed to within the business logic of the client.
  Here are some sample use cases:

  *Use Case 1: Automatically populate URL protocol*
  ```python
  ignition.url('//gemini.circumlunar.space') # => gemini://gemini.circumlunar.space
  ```

  *Use Case 2: Navigate to an absolute path*
  ```python
  ignition.url('/home', 'gemini://gemini.circumlunar.space') # => gemini://gemini.circumlunar.space/home
  ```

  *Use Case 3: Navigate to a relative path*
  ```python
  ignition.url('2', 'gemini://gemini.circumlunar.space/home') # => gemini://gemini.circumlunar.space/home/2
  ```

  *Use Case 4: Resolve paths with navigation*
  ```python
  ignition.url('../fun/', 'gemini://gemini.circumlunar.space/home/work/') # => gemini://gemini.circumlunar.space/home/fun/
  ```

  *Note:* if the user's intent is to generate a url to a Gemini capsule and then make a request,
  ignition recommends that you just provide the *url* and *referer* to `ignition.request()`, as
  that function encapsulates all of the logic within this method when making a request.  If you
  want to retrieve a URL from an already processed request, it is recommended to use
  `ignition.BaseResponse.url`, as that will store the URL that was actually used.  This method
  is only intended for use in constructing a URL but not generating a request.

  Parameters:
  * url: `string`
  * referer: `string` (optional)

  Returns: `string`
  '''
  dummy_req = Request(request_url, referer=referer)
  return dummy_req.get_url()


def request(request_url, referer=None, timeout=None, ca_cert=None):
  '''
  Given a *url* to a Gemini capsule, this performs a request to the specified
  url and returns a response (as a subclass of [ignition.BaseResponse](#ignitionbaseresponse))
  with the details associated to the response.  This is the interface that most
  users should use.

  If a *referer* is provided, a dynamic URL is constructed by ignition to send a
  request to. (*referer* expectes a fully qualified url as returned by
  `ignition.BaseResponse.url` or (less prefered) `ignition.url()`).
  Typically, in order to simplify the browsing experience, you should pass
  the previously requested URL as the referer to simplify URL construction logic.

  *See `ignition.url()` for details around url construction with a referer.*

  If a *timeout* is provided, this will specify the client timeout (in seconds)
  for this request.  The default is 30 seconds.  See also `ignition.set_default_timeout`
  to change the default timeout.

  If a *ca_cert* is provided, the certificate will be sent to the server as a CA CERT.
  You will need to provide the paths to both the certificate and the key in this case.

  Depending on the response from the server, as per Gemini specification, the
  corresponding response type will be returned.

  * If the response status begins with "1", the response type is `INPUT`,
    and will return a response of type [ignition.InputResponse](#ignitioninputresponse).
  * If the response status begins with "2", the response type is `STATUS`,
    and will return a response of type [ignition.SuccessResponse](#ignitionsuccessresponse).
  * If the response status begins with "3", the response type is `REDIRECT`,
    and will return a response of type [ignition.RedirectResponse](#ignitionredirectresponse).
  * If the response status begins with "4", the response type is `TEMPORARY FAILURE`,
    and will return a response of type [ignition.TempFailureResponse](#ignitiontempfailureresponse).
  * If the response status begins with "5", the response type is `PERMANENT FAILURE`,
    and will return a response of type [ignition.PermFailureResponse](#ignitionpermfailureresponse).
  * If the response status begins with "6", the response type is `CLIENT CERTIFICATE REQUIRED`,
    and will return a response of type [ignition.ClientCertRequiredResponse](#ignitionclientcertrequiredresponse).
  * If *no valid response* can be returned, ignition assigns a response type of "0"
    and returns a response of type [ignition.ErrorResponse](#ignitionerrorresponse).

  Parameters:
  * url: `string`
  * referer: `string` (optional)
  * timeout: `float` (optional)
  * ca_cert: `Tuple(cert_file, key_file)` (optional)

  Returns: `[ignition.BaseResponse](#ignitionbaseresponse)`
  '''

  req = Request(request_url, cert_store=__cert_store, request_timeout=__timeout.get_timeout(timeout), referer=referer, ca_cert=ca_cert)

  return req.send()
